import datetime
from time import time
from importlib.metadata import version
import logging


def log(msg, level=logging.NOTSET):
    logging.getLogger('gents').log(level=level, msg=msg)

def get_time_stamp():
    return datetime.datetime.fromtimestamp(time()).strftime('%Y-%m-%d %H:%M')

def get_version():
    return version('gents')


#!/usr/bin/env python
"""
timeseries.py

Developer: Cameron Cummins
Contact: cameron.cummins@utexas.edu
Last Header Update: 01/31/25
"""
import netCDF4
import numpy as np
from os.path import isfile
from os import remove, makedirs
from cftime import num2date
from gents.meta import get_attributes


def is_var_secondary(variable: netCDF4._netCDF4._Variable,
                     secondary_vars: list = ["time_bnds", "time_bnd", "time_bounds", "time_bound"],
                     secondary_dims: list = ["nbnd", "chars", "string_length", "hist_interval"],
                     max_num_dims: int = 1,
                     primary_dims: list = ["time"]) -> bool:
    """
    Determines if a variable is secondary or not (and then should be included in all time series files).
    Criteria are applied in order of the parameters described.

    :param dimensions: Dimensions for the variable.
    :param secondary_dims: List of secondary variable names
    :param secondary_dims: List of dimensions that make a variable secondary
    :param max_num_dims: Maximum number of dimensions a secondary variable should have.
    :param primary_dims: List of dimensions that make a variable primary (thus not secondary).
    :return: True if the variable is secondary, false if not.
    """
    if variable.name in secondary_vars:
        return True
        
    dims = np.unique(variable.dimensions)

    for tag in secondary_dims:
        if tag in dims:
            return True

    if len(dims) > max_num_dims:
        for tag in primary_dims:
            if tag in dims:
                return False

    return True


def get_timestamp_str(times):
    """
    Creates timestamp string to describe time range for netCDF dataset

    :param times: Time values for netCDF dataset in integer form with units and calendar attributes.
    :return: String containing appropriate timestamp.
    """
    start_t = num2date(times[0], units=times.units, calendar=times.calendar)
    if times.shape[0] == 1:
        return start_t.strftime("%Y-%m-%d-%H:%M:%S")
    else:
        dt = num2date(times[1], units=times.units, calendar=times.calendar) - start_t
        minutes = dt.total_seconds() / 60
        hours = minutes / 60
        days = hours / 24
        months = days / 30
    
        if minutes < 1:
            time_format = "%Y-%m-%d-%H:%M:%S"
        elif hours < 1:
            time_format = "%Y-%m-%d-%H"
        elif days < 1:
            time_format = "%Y-%m-%d"
        elif months < 1:
            time_format = "%Y-%m"
        else:
            time_format = "%Y"
        
        end_t = num2date(times[-1], units=times.units, calendar=times.calendar)
        return f"{start_t.strftime(time_format)}.{end_t.strftime(time_format)}"


def generate_time_series(hf_paths, ts_out_dir, prefix=None, complevel=0, compression=None, overwrite=False):
    """
    Creates timeseries dataset from specified history file paths.

    :param hf_paths: List of paths to history files to generate time series from
    :param ts_out_dir: Directory to output time series files to.
    :param prefix: Prefix to add to beginning of the names of the generated time series files.
    :param complevel: Compression level to apply through netCDF4 API.
    :param compression: Compression algorithm to use through netCDF4 API.
    :param overwrite: Whether or not to delete existing time series files with the same names as those being generated.
    :return: List of paths to time series generated.
    """
    agg_hf_ds = netCDF4.MFDataset(hf_paths, aggdim="time")
    
    global_attrs = get_attributes(agg_hf_ds)
    primary_vars = []
    secondary_vars_data = {}
    output_paths = []
    
    for variable in agg_hf_ds.variables:
        if is_var_secondary(agg_hf_ds[variable]):
            secondary_vars_data[variable] = agg_hf_ds[variable][:]
        else:
            primary_vars.append(variable)
    
    for variable in primary_vars:
        ts_string = get_timestamp_str(agg_hf_ds["time"])
        if prefix is not None:
            ts_out_path = f"{ts_out_dir}/{prefix}.{variable}.{ts_string}.nc"
        else:
            ts_out_path = f"{ts_out_dir}/{variable}.{ts_string}.nc"

        if not overwrite and isfile(ts_out_path):
            # Can add a check here to see if the file is readable and the version attribute is added
            continue
        elif overwrite and isfile(ts_out_path):
            remove(ts_out_path)

        if not ts_out_dir.exists():
            makedirs(ts_out_dir)
        
        var_ds = agg_hf_ds[variable]
        ts_ds = netCDF4.Dataset(ts_out_path, mode="w")
        
        for index, dim in enumerate(var_ds.dimensions):
            if dim == "time":
                ts_ds.createDimension(dim, None)
            else:
                ts_ds.createDimension(dim, var_ds.shape[index])
    
        var_data = ts_ds.createVariable(variable,
                                        var_ds.dtype,
                                        var_ds.dimensions,
                                        complevel=complevel,
                                        compression=compression)
        var_data.set_auto_mask(False)
        var_data.set_auto_scale(False)
        var_data.set_always_mask(False)
        
        ts_ds[variable].setncatts(get_attributes(var_ds))
    
        time_chunk_size = 1
        if len(var_ds.shape) > 0 and "time" in var_ds.dimensions:
            for i in range(0, var_ds.shape[0], time_chunk_size):
                if i + time_chunk_size > var_ds.shape[0]:
                    time_chunk_size = var_ds.shape[0] - i
                var_data[i:i + time_chunk_size] = var_ds[i:i + time_chunk_size]
        else:
            var_data[:] = var_ds[:]
    
        for secondary_var in secondary_vars_data:
            svar_ds = agg_hf_ds[secondary_var]
            
            for index, dim in enumerate(svar_ds.dimensions):
                if dim not in ts_ds.dimensions:
                    if dim == "time":
                        ts_ds.createDimension(dim, None)
                    else:
                        ts_ds.createDimension(dim, svar_ds.shape[index])
            
            svar_data = ts_ds.createVariable(secondary_var,
                                             svar_ds.dtype,
                                             svar_ds.dimensions,
                                             complevel=complevel,
                                             compression=compression)
            
            svar_data.set_auto_mask(False)
            svar_data.set_auto_scale(False)
            svar_data.set_always_mask(False)
    
            ts_ds[secondary_var].setncatts(get_attributes(svar_ds))
            svar_data[:] = secondary_vars_data[secondary_var]
        
        # Put version here later
        ts_ds.setncatts(global_attrs | {"gents_version": "put version here later"})
        ts_ds.close()
        output_paths.append(ts_out_path)
    return output_paths


#!/usr/bin/env python
"""
read.py

Developer: Cameron Cummins
Contact: cameron.cummins@utexas.edu
Last Header Update: 01/28/25
"""
from gents.utils import log
from gents.meta import get_meta_from_path
from cftime import num2date
from pathlib import Path
import numpy as np
import os
import fnmatch
import cftime
import netCDF4
import dask


def is_ds_within_years(ds_meta, min_year, max_year):
    time_bounds = ds_meta.get_cftime_bounds()[0]
    year = (time_bounds[0].year + time_bounds[1].year) / 2

    if min_year <= year <= max_year:
        return True
    else:
        return False


def apply_filter(meta_ds, filter_dict):
    if "year_bounds" in filter_dict:
        min_year, max_year = filter_dict["year_bounds"]
        if is_ds_within_years(meta_ds, min_year, max_year):
            return True
    return False


def apply_inclusive_filters(path_meta_map, filters):
    filtered_mapping = {}

    for path in path_meta_map:
        meta_ds = path_meta_map[path]
    
        for tag in filters:
            if tag in str(path) and apply_filter(meta_ds, filters[tag]):
                filtered_mapping[path] = meta_ds
                break
    return filtered_mapping


def apply_exclusive_filters(path_meta_map, filters):
    filtered_mapping = {}
    for path in path_meta_map:
        tag_found = False
        for tag in filters:
            if tag in str(path):
                tag_found = True
                break

        if not tag_found:
            filtered_mapping[path] = path_meta_map[path]
        
    return filtered_mapping


def check_config(config):
    assert "name" in config
    assert "include" in config
    assert "exclude" in config
    assert type(config["include"]) is dict or config["include"] is None
    assert type(config["exclude"]) is list or config["exclude"] is None


def get_default_config():
    return {
        "name": "default",
        "include": None,
        "exclude": None
    }


def find_files(head_path, pattern):
    """
    Search for files in the specified head directory and all subdirectories that match the given wildcard pattern.

    :param head_path: The head directory to start searching from.
    :param pattern: The wildcard pattern to match files against (e.g., '*.nc').
    :return: A list of file paths that match the pattern.
    """
    matched_files = []

    for root, dirs, files in os.walk(head_path):
        for file in files:
            if fnmatch.fnmatch(file, pattern):
                matched_files.append(Path(os.path.join(root, file)))

    return matched_files


def calculate_year_slices(slice_size_years, min_year, max_year):
    """
    Calculates the ranges for each slice in years within a given range.

    :param slice_size_years: Length of each slice in years.
    :param min_year: Minimum or starting year for the full range.
    :param max_year: Maximum or ending year for the full range.
    :return: List of tuples where each tuple defines the year range for each slice
    """
    start_year = int(np.floor(min_year / slice_size_years)*slice_size_years)
    end_year = int(np.ceil(max_year / slice_size_years)*slice_size_years)

    ranges = []
    for year in np.arange(start_year, end_year, slice_size_years, dtype=int):
        ranges.append((int(year), int(year+slice_size_years-1)))
    
    return ranges


def find_all_indices(string, substring):
    """
    Finds all indices where the substring occurs in a string.

    :param string: The string to search in.
    :param substring: The substring to search for.
    :return: A list of indices where the substring is found.
    """
    indices = []
    start = 0
    while True:
        index = string.find(substring, start)
        if index == -1:
            break
        indices.append(index)
        start = index + 1
    return indices


def sort_hf_groups(hf_paths, delimiter=".", substring_index=2):
    """
    Groups history file paths by directory and substrings.

    :param hf_paths: List of paths to history files.
    :param delimiter: The delimiter to find the substrings with.
    :param substring_index: Substrings to skip after splitting with the delimiter (from right to left).
    :return: A dictionary that maps group substrings (key) to history file paths (value)
    """
    directory_groups = {}
    for path in hf_paths:
        if path.parent in directory_groups:
            directory_groups[path.parent].append(path)
        else:
            directory_groups[path.parent]= [path]

    hf_groups = {}
    for parent_path in directory_groups:
        group_paths = [path for path in directory_groups[parent_path]]
        substrings = []
        for path in group_paths:
            parsed = path.name[:find_all_indices(path.name, delimiter)[-1 * substring_index]]
            substrings.append(parsed)
        
        for substring in np.unique(substrings):
            hf_groups[f"{parent_path}/{substring}"] = []
            for path in group_paths:
                parsed = path.name[:find_all_indices(path.name, delimiter)[-1 * substring_index]]
                if substring == parsed:
                    hf_groups[f"{parent_path}/{substring}"].append(path)
        
    return hf_groups


def get_year_bounds(hf_to_meta_map):
    """
    Determines the minimum and maximum year for a series of mapped history files.

    :param hf_to_time_map: Dictionary that maps history file paths (key) to their time and time bound data (value)
    :return: The minimum and maximum years from the specified history files as a tuple
    """
    min_year = np.inf
    max_year = -np.inf
    
    for path in list(hf_to_meta_map.keys()):
        time_bounds = hf_to_meta_map[path].get_cftime_bounds()
        if time_bounds is None:
            time_bounds = []
            for ts in hf_to_meta_map[path].get_cftimes():
                time_bounds.append([ts, ts])
        
        for index in range(len(time_bounds)):
            lower_bound, upper_bound = time_bounds[index]
            
            mid_ordinal = np.median([lower_bound.toordinal(), upper_bound.toordinal()])
            mid_time = cftime.datetime.fromordinal(mid_ordinal, calendar=lower_bound.calendar)
                    
            if mid_time.year > max_year:
                max_year = mid_time.year
            if mid_time.year < min_year:
                min_year = mid_time.year
    return min_year, max_year


def slice_hf_groups(hf_groups, hf_to_meta_map, slice_size_years):
    """
    Seperates history files path groups into sliced time chunks.

    :param hf_groups: A dictionary that maps group substrings (key) to history file paths (value), generated by sort_hf_groups
    :param hf_to_meta_map: Dictionary that maps history file paths (key) to respective metadata (value)
    :param slice_size_years: Number of years per slice.
    :return: 
    """
    sliced_groups = {}
    for group in hf_groups:
        hf_paths = hf_groups[group]
        if len(hf_paths) == 1:
            # Probably good to throw a warning here
            continue
        else:
            group_meta_map = {path: hf_to_meta_map[path] for path in hf_paths}
            
            min_year, max_year = get_year_bounds(group_meta_map)
            time_slices = calculate_year_slices(slice_size_years, min_year, max_year+1)
            time_slices[0] = (min_year, time_slices[0][1])
            time_slices[-1] = (time_slices[-1][0], max_year)
            
            hf_slices = {}
            variable_set = None
            for hf_path in hf_paths:
                meta_ds = hf_to_meta_map[hf_path]
                time = meta_ds.get_cftimes()
                time_bnds = meta_ds.get_cftime_bounds()
                variables = meta_ds.get_variables()

                if time is not None:
                    time = time[0]
                
                if time_bnds is not None:
                    time_bnds = time_bnds[0]
                    time = time_bnds[0] + (time_bnds[1] - time_bnds[0]) / 2
                
                for time_slice in time_slices:
                    if time_slice[0] <= time.year <= time_slice[1]:
                        if time_slice in hf_slices:
                            hf_slices[time_slice].append(meta_ds)
                        else:
                            hf_slices[time_slice] = [meta_ds]
                        break
    
            for time_slice in hf_slices:
                sliced_groups[f"{group}.{time_slice[0]}-{time_slice[1]}"] = hf_slices[time_slice]
    return sliced_groups


def generate_output_template(hf_head_dir, group_path_id, output_head_dir=None, directory_swaps={"hist": "tseries"}, filename_delimiter=".", cutoff_index=None):
    """
    Creates file path template (missing file name suffix) for outputing timeseries files to.

    :param hf_head_dir: Head directory used to read in the history files.
    :param group_path_id: One group path ID produced by slice_hf_groups.
    :param output_head_dir: Head directory to output to, if not the same as the input head directory.
    :param directory_swaps: Dictionary with directory names (key) to rename (value)
    :param filename_delimiter: Delimiter used to separate file name tags
    :param cutoff_index: Index to cutoff tags that are generated using the delimiter (typically used to remove the date)
    :return: Output template path for generating time series files.
    """
    group_path_id = Path(group_path_id)
                         
    raw_filename_prefix = group_path_id.name
    if cutoff_index is None:
        cutoff_index = find_all_indices(raw_filename_prefix, ".")[-1]
    filename_prefix = raw_filename_prefix[:cutoff_index]
    
    sub_dir_structure = (str(group_path_id.parent).split(hf_head_dir)[-1]).split("/")

    for key in directory_swaps:
        for index in range(len(sub_dir_structure)):
            if sub_dir_structure[index] == key:
                sub_dir_structure[index] = directory_swaps[key]

    sub_dir_path = "/"
    for directory in sub_dir_structure:
        sub_dir_path += f"{directory}/"

    if output_head_dir is None:
        output_template = Path(f"{hf_head_dir}/{sub_dir_path}/{filename_prefix}")
    else:
        output_template = Path(f"{output_head_dir}/{sub_dir_path}/{filename_prefix}")
    return output_template


def filter_by_variables(meta_datasets):
    variable_sets = {}
    for index in range(len(meta_datasets)):
        var_set = meta_datasets[index].get_variables()
        var_set.sort()

        if str(var_set) in variable_sets:
            variable_sets[str(var_set)].append(index)
        else:
            variable_sets[str(var_set)] = [index]

    majority = None
    others = None
    
    if len(variable_sets) > 1:
        counts = []
        for var_set in variable_sets:
            counts.append(len(variable_sets[var_set]))

        majority_index = np.argmax(counts)
        if np.sum(counts[majority_index] == np.array(counts)) == 1:
            majority_set = list(variable_sets)[majority_index]
    
            majority = []
            others = []
            for index in range(len(meta_datasets)):
                if index in variable_sets[majority_set]:
                    majority.append(meta_datasets[index])
                else:
                    others.append(meta_datasets[index])
    else:
        majority = meta_datasets
    
    return (majority, others)


def sort_metas_by_time(metas):
    time_sorted_metas = [metas[0]]

    if len(metas) > 1:
        for meta in metas[1:]:
            meta_sorted = False
            
            for index in range(len(time_sorted_metas)):
                if meta.get_cftimes()[0] < time_sorted_metas[index].get_cftimes()[0]:
                    time_sorted_metas.insert(index, meta)
                    meta_sorted = True
                    break
            if not meta_sorted:
                time_sorted_metas.append(meta)

    return time_sorted_metas

    
def check_groups_by_variables(sliced_groups):
    filtered_sliced_groups = {}
    for group in sliced_groups:
        meta_datasets = sliced_groups[group]
        majority, others = filter_by_variables(meta_datasets)
        if majority is not None:
            filtered_sliced_groups[group] = sort_metas_by_time(majority)
            if others is not None:
                for meta_ds in others:
                    log(f"Dataset has inconsistent variable list with directory group: {meta_ds.get_path()}")
        else:
            log(f"Unable to determine majority dataset, check variable configurations between directory groups, group ID: {group}")
    return filtered_sliced_groups


def get_metas_from_paths(paths, dask_client=None):
    if dask_client is None:
        dask_client = dask.distributed.client._get_global_client()
        
    ds_metas_futures = dask_client.map(get_meta_from_path, paths)
    ds_metas = dask_client.gather(ds_metas_futures)
    del ds_metas_futures
    
    hf_to_meta_map = {path: ds_metas[index] for index, path in enumerate(paths) if ds_metas[index] is not None and ds_metas[index].get_cftime_bounds() is not None}
    return hf_to_meta_map


def get_groups_from_paths(paths, slice_size_years=10, dask_client=None):
    if dask_client is None:
        dask_client = dask.distributed.client._get_global_client()

    hf_to_meta_map = get_metas_from_paths(paths, dask_client=dask_client)
    
    groups = sort_hf_groups(list(hf_to_meta_map.keys()))
    sliced_groups = slice_hf_groups(groups, hf_to_meta_map, slice_size_years)
    filtered_sliced_groups = check_groups_by_variables(sliced_groups)

    return filtered_sliced_groups


#!/usr/bin/env python
"""
gents_cli

Command Line Interface for running GenTS in the terminal without a notebook.

Developer: Cameron Cummins
Contact: cameron.cummins@utexas.edu
Last Header Update: 8/23/24
"""
import gents
import argparse
from dask.distributed import Client
from json import load
from importlib.metadata import version as getVersion


def main():
    parser = argparse.ArgumentParser(description = f"Command Line Interface (CLI) for Generating Time-Series (GenTS) Version {getVersion("gents")}")

    parser.add_argument("hf_head_dir", nargs = 1, metavar = "hf_head_dir", type = str, 
                         help = "Path to head directory with subdirectories containing history files.")
    parser.add_argument("ts_head_dir", nargs = 1, metavar = "ts_head_dir", type = str, 
                         help = "Path to head directory where structure reflecting `hf_head_dir` will be created and timeseries files will be written to.")
    parser.add_argument("-d", "--dask_address", nargs = "?", metavar = "dask_address", type = str, 
                         help = "Address to dask scheduler managing an active dask cluster.")
    parser.add_argument("-s", "--dir_name_swaps_path", nargs = "?", metavar = "dir_name_swaps_path", type = str, 
                         help = "Path to JSON containing dictionary for swapping out keyword directory names in the structure under `hf_head_dir` (e.g. ``{'hist' : 'proc/tseries'}``")
    parser.add_argument("-F", "--file_exclusions", nargs = '*', metavar = "file_exclusions", type = str, 
                         help = "File names containing any of the keywords in this list will be excluded from the database (Default: ['rest', 'logs']).")
    parser.add_argument("-D", "--dir_exclusions", nargs = '*', metavar = "dir_exclusions", type = str, 
                         help = "Directory names containing any of the keywords in this list will be excluded from the database (Default: ['rest', 'logs']).")
    parser.add_argument("-l", "--timeseries_year_length", nargs = "?", metavar = "timeseries_year_length", type = int, 
                         help = "Number of years each timeseries file should be chunked to using `getYearSlices` (Default: 10).")
    parser.add_argument("-i", "--include_variables", nargs = '*', metavar = "include_variables", type = str, 
                         help = "Variables to include in either creating individual timeseries files for adding as auxiliary variables (Default: None).")
    parser.add_argument("-e", "--exclude_variables", nargs = '*', metavar = "exclude_variables", type = str, 
                         help = "Variables to exclude from either creating individual timeseries files for adding as auxiliary variables (Default: None).")
    parser.add_argument("-n", "--year_start", nargs = "?", metavar = "year_start", type = int, 
                         help = "Starting year for timeseries generation, must be later than first history file timestamp to have an effect (Default: None).")
    parser.add_argument("-m", "--year_end", nargs = "?", metavar = "year_end", type = int, 
                         help = "Ending year for timeseries generation, must be later than last history file timestamp to have an effect (Default: None).")
    parser.add_argument("-C", "--compression_level", nargs = "?", metavar = "compression_level", type = int, 
                         help = "Compression level to pass to netCDF4 engine when generating timeseries files (Default: None).")
    parser.add_argument("-a", "--compression_algo", nargs = "?", metavar = "compression_algo", type = str, 
                         help = "Compression algorithm to pass to netCDF4 engine when generating timeseries files. See netCDF4-python documentation for available algorithms (Default: 'bzip2').")
    parser.add_argument("-c", "--variable_compression_levels_path", nargs = "?", metavar = "variable_compression_levels_path", type = str, 
                         help = "Path to JSON containing compression levels to apply to specific variables. Variable name is key and the compression level is the value (Default: None).")
    parser.add_argument("-v", "--verbosity_level", nargs = "?", metavar = "verbosity_level", type = int, 
                         help = "Level of logging to output to the standard output stream. 0 = No output, 1 = high level computational stages with timings, 2 = all stages and iterations with timings, including logic (Default: 1).")
    parser.add_argument("--overwrite", action='store_true', 
                         help = "Overwrite timeseries files if they already exist at the generated paths.")
    args = parser.parse_args()

    if args.hf_head_dir is not None:
        hf_head_dir = args.hf_head_dir[0]

    if args.ts_head_dir is not None:
        ts_head_dir = args.ts_head_dir[0]

    if args.dask_address is not None:
        client = Client(args.dask_address[0])
    else:
        client = None

    if args.dir_name_swaps_path is not None:
        with open(args.dir_name_swaps_path[0], "r") as f:
            variable_compression_levels = load(f)
    else:
        dir_name_swaps = {}

    if args.file_exclusions is not None:
        file_exclusions = args.file_exclusions
    else:
        file_exclusions = []

    if args.dir_exclusions is not None:
        dir_exclusions = args.dir_exclusions
    else:
        dir_exclusions = []

    if args.timeseries_year_length is not None:
        timeseries_year_length = args.timeseries_year_length[0]
    else:
        timeseries_year_length = 10

    if args.include_variables is not None:
        include_variables = args.include_variables
    else:
        include_variables = None

    if args.exclude_variables is not None:
        exclude_variables = args.exclude_variables
    else:
        exclude_variables = None

    if args.year_start is not None:
        year_start = args.year_start[0]
    else:
        year_start = None

    if args.year_end is not None:
        year_end = args.year_end[0]
    else:
        year_end = None

    if args.compression_level is not None:
        compression_level = args.compression_level[0]
    else:
        compression_level = None

    if args.compression_algo is not None:
        compression_algo = args.compression_algo[0]
    else:
        compression_algo = None

    if args.variable_compression_levels_path is not None:
        with open(args.variable_compression_levels_path[0], "r") as f:
            variable_compression_levels = load(f)
    else:
        variable_compression_levels = None

    if args.verbosity_level is not None:
        verbosity_level = args.verbosity_level[0]
    else:
        verbosity_level = 1

    if args.overwrite is not None:
        overwrite = args.overwrite
    else:
        overwrite = False


    modb = gents.ModelOutputDatabase(
        hf_head_dir=hf_head_dir,
        ts_head_dir=ts_head_dir,
        dir_name_swaps=dir_name_swaps,
        file_exclusions=file_exclusions,
        dir_exclusions=dir_exclusions,
        timeseries_year_length=timeseries_year_length,
        overwrite=overwrite,
        include_variables=include_variables,
        exclude_variables=exclude_variables,
        year_start=year_start,
        year_end=year_end,
        compression_level=compression_level,
        compression_algo=compression_algo,
        variable_compression_levels=variable_compression_levels,
        verbosity_level=verbosity_level
    )

    modb.run(client=client)


if __name__ == '__main__':
    main()